#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Transitive clojure"""


def strinset(node1, node2, rel):
    if rel in {"r", "dn", "dp"}:
        return str(node1)+"|"+str(node2)+"|"+rel
    elif rel == "up":
        return str(node2)+"|"+str(node1)+"|"+"dn"
    elif rel in {"sv", "sn", "o"}:
        if node1 < node2:
            return str(node1)+"|"+str(node2)+"|"+rel
        else:
            return str(node2)+"|"+str(node1)+"|"+rel
    else:
        raise TypeError

def comprel(rel1, rel2):
    if   (rel1 == "r") and (rel2 == "r"):
        return "r", False
    elif (rel1 == "r") and (rel2 == "dn"):
        return "r", False
    elif (rel1 == "up") and (rel2 == "r"):
        return "r", False
    elif (rel1 == "up") and (rel2 == "up"):
        return "up", False
    elif (rel1 == "dn") and (rel2 == "dn"):
        return "dn", False
    elif (rel1 == "sv") and (rel2 == "sv"):
        return "sv", False
    elif (rel1 == "sn") and (rel2 == "sn"):
        return "sn", False
    elif (rel1 in {"sn", "sv"}) and (rel2 == "dn"):
        return "dn", False
    elif (rel1 in {"sn", "sv"}) and (rel2 == "up"):
        return "up", False
    elif (rel1 == "dn") and (rel2 in {"sn", "sv"}):
        return "dn", False
    elif (rel1 == "up") and (rel2 in {"sn", "sv"}):
        return "up", False
    elif (rel1 == "sn") and (rel2 == "sv"):
        return "sn", True
    elif (rel1 == "sv") and (rel2 == "sn"):
        return "sv", True
    elif (rel1 == "ref"):          # at begining and accept every relation include overlap and depends-on
        return rel2, False
    else:
        return None, False

def disjoint_head(fa, i):
    if i not in fa.keys():
        fa[i] = i
        return i
    else:
        while fa[fa[i]] != fa[i]:
            fa[i] = fa[fa[i]]
        return fa[i]

def get_cluster(graph):

    """
    The existence of non-transitivity makes it necessary to start a DFS from every node, for example, a coref go to superset and go to another subset, this does not hold for transitive anymore, but it is a cluster. The cluster is a method to avoid repetitive counting of any nodes.
    If a node does not have a cluster yet, it will be assigned a new cluster. As long as it reaches an existing cluster, its current cluster will be assimilated by the existing cluster, and the cluster size will increase by one, and this node will be assigned to that existing cluster.
    But we'll leave this step to the final step. Now each node will form a separate cluster itself.

    There is also another problem with dealing with cycles: if there are cycles, if the relation is not the same as before, then there is inconsistency.
    """

    node2cluster = {}
    stack = []
    fa = {}
    whole_set = set()
    inconsistant_pairs = set()
    circular_nodes = set()
    wrong_coref_triples = set()

    for start_node in graph.keys():

        stack.append((start_node, "ref"))
        this_cluster = set()
        visited_nodes = set()
        if start_node not in fa:          # init
            fa[start_node] = start_node
        else:
            fa[start_node] = disjoint_head(fa, start_node)
        while stack:
            now, nowrel = stack.pop()
            visited_nodes.add(now)
            # all nodes visited should be marked as visited, including terminal nodes
            for next_node, next_rel in graph[now]:
                # the trick here is to refresh the disjoint head of all nodes to the new one
                fa[disjoint_head(fa, next_node)] = fa[start_node]                         # we must change its head, not itself. Any connected node will be considered in the same cluster, but not necessarily transitive
                updated_rel = comprel(nowrel, next_rel)
                if updated_rel[1]:
                    wrong_coref_triples.add((start_node, now, next_node))
                if updated_rel[0]:                                                     # ensure that the next node is viable from current standing
                    # it should be that: visited nodes relations are recorded but not added to stack
                    if (next_node not in visited_nodes):
                        this_cluster.add(strinset(start_node, next_node, updated_rel[0]))   # add to the relations
                        if next_node in graph.keys():                        # if a node is the final node of a chain
                            stack.append((next_node, updated_rel[0])) # later it will also be added
                        else:
                            visited_nodes.add(next_node)
                    else:
                        # next_node visited, need to check whether contradictory
                        if (start_node != next_node):
                            if (strinset(start_node, next_node, updated_rel[0]) not in this_cluster):
                            # print(f"Inconsistency detected in node pair {start_node} and {next_node}")
                                inconsistant_pairs.add((start_node, next_node))
                                this_cluster.add(strinset(start_node, next_node, updated_rel[0]))
                                # or else they should have the same relationship
                        else:
                            if updated_rel[0] not in {"sn", "sv", "ref"}:
                                # print(f"Circular relations in node {start_node}!")
                                circular_nodes.add((start_node, ))
                                this_cluster.add(strinset(start_node, next_node, updated_rel[0]))

            # the reason why we need to loop through all nodes is some nodes after "overlap" or "depends-on" are neither in visited_nodes nor updated, and if they are used next time, their information should serve as gold, which is hazadous; if they are not looped through, ultimately they need also be updated

        node2cluster[start_node] = [this_cluster, visited_nodes]

    for node in fa.keys():
        if fa[fa[node]] != fa[node]:
            fa[node] = disjoint_head(fa, node)

    for node in graph.keys():
        whole_set.update(node2cluster[node][0])
        if fa[node] != node:
            node2cluster[fa[node]][0].update(node2cluster[node][0])
            node2cluster[fa[node]][1].update(node2cluster[node][1])
            del node2cluster[node]

    return node2cluster.values(), whole_set, inconsistant_pairs, circular_nodes, wrong_coref_triples



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Various operations"""

import numpy as np
from ancast.params import SENSE_COEFFICIENT


def string_sim(str1a, str2a):
    str1 = str1a.lower()
    str2 = str2a.lower()

    if (str1 not in str2) and (str2 not in str1):
        return 0.0
    elif str1 in str2:
        return len(str1) / len(str2)
    elif str2 in str1:
        return len(str2) / len(str1)
    else:
        raise NotImplementedError

def jaccard_sim(set1, set2):
    intersection = set1.intersection(set2)
    union = set1.union(set2)
    jaccard_index = len(intersection) / len(union) if union else 0
    return jaccard_index

def name_sim(str1, str2, sense_sim, sense_coefficient=SENSE_COEFFICIENT):
    return string_sim(str1, str2) * (1 + sense_coefficient * (sense_sim - 1))

def rel_weight(a, b):
    return np.sqrt(a*b) + 1

def get_argcnt(edge_matrix, node):
    node_set = set()
    q = [node]
    while q:
        c = q.pop(0)
        nexts = np.where(edge_matrix[c] > 0)
        for i in nexts[0].tolist():
            if (i not in node_set):
                node_set.add(i)
                q.append(i)
    return len(node_set)

def get_children(edge_matrix, from_node, to_node):
    # traverse, one node counted twice
    from_node_set = set()
    to_node_set = set()
    q = [from_node]
    while q:
        c = q.pop(0)
        nexts = np.where(edge_matrix[c] > 0)
        for i in nexts:
            if (i not in from_node_set) and (i != to_node):
                from_node_set.add(i)
                q.append(i)
        q = [from_node]
    q = [to_node]
    while q:
        c = q.pop(0)
        nexts = np.where(edge_matrix[c] > 0)
        for i in nexts:
            if (i not in to_node_set):
                to_node_set.add(i)
                q.append(i)
    return len(from_node_set), len(to_node_set)

def protected_divide(a, b):
    if b==0:
        return 0.0
    else:
        return a / b

def parse_alignment(alignment_text):
    ras = {}
    if not alignment_text:
        return ras
    als = alignment_text.strip().split("\n")
    for l in als:
        if "alignment" in l:
            continue
        v, al = l.strip().split(":")
        ras[v] = al.strip()
    return ras

# def rel_hash(word1, rel, word2):
#     return word1.var+"-"+rel+"-"+word2.var, word2.var+"-"+reverse(rel)+"-"+word1.var

# ======================== relation ===========================
def reverse(rel):
    if rel[-3:] == "-of":
        return rel[:-3]
    else:
        return rel + "-of"

def is_reverse(rel):
    if rel == "mod":        # mod is "domain-of"
        return True
    elif rel == "mod-of":
        return False
    elif rel[-3:] == "-of":
        return True
    else:
        return False

# ================= for label match =================
def rel_label_set_comp(rel_label_set, target_label_set):
    """
    For possible fine-grained match.
    """
    set_a = rel_label_set
    set_b = target_label_set
    return len(set_a & set_b) # simply count fully equal labels



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""resources"""

import importlib.resources
import json

ABT_VOCAB: dict = {}
REIFY_RELS: dict = {}
ALLOWED_TAGS: set = {"explicit"}

# resources
def load_resource(fname: str):
    return json.loads(
        importlib.resources.read_text('ancast.resources', fname)
    )

def load_abt_vocab(fname='abstract_vocab.json', global_cache=True):
    abt_vocab = load_resource(fname)
    if global_cache:
        global ABT_VOCAB
        ABT_VOCAB = abt_vocab
    return abt_vocab

def load_reify_rels(fname='reification_relations.json', global_cache=True):
    reify_rels = load_resource(fname)
    if global_cache:
        global REIFY_RELS
        REIFY_RELS = reify_rels
        ALLOWED_TAGS.add('reified')
    return reify_rels



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Word representation"""

import logging
from collections import defaultdict

from ancast.node import Attribute, IMPLICIT
from ancast.ops import reverse, is_reverse
from ancast.resource_utils import ALLOWED_TAGS, ABT_VOCAB

logger = logging.getLogger(__name__)


def add_vocab(vocab, name, sense_id, sense):
    if name not in vocab.keys():
        vocab[name] = {}

    if sense_id in vocab[name].keys():
        if vocab[name][sense_id] != sense:
            print(name)
            print("Original ", vocab[name])
            print("New ",sense)
            sense_id += 10

    vocab[name][sense_id] = sense


# local vocab (TODO: rework)
VOCAB = {}

class Word:

    def __init__(self, raw_name, var, sense_id = 0):

        self.raw_name = raw_name
        self.name = raw_name.lower()
        self.var = var
        self.sense_id = sense_id
        self.relations = defaultdict(list)

        if self.name in VOCAB:
            try:
                self.sense = VOCAB[self.name][sense_id]
            except KeyError:
                # where a noun and a verb has the same surface form whereas the verb's propbank does not consider its noun form
                add_vocab(vocab= VOCAB, name= self.name, sense_id = self.sense_id, sense= {})
                self.sense = {}
            self.v = VOCAB
        elif self.name in ABT_VOCAB:
            try:
                self.sense = ABT_VOCAB[self.name][sense_id]
            except KeyError:
                # sometimes there are errors in annotating, like have-degree-91, which should be 92.
                add_vocab(vocab= ABT_VOCAB, name= self.name, sense_id = self.sense_id, sense= {})
                self.sense = {}
            self.v = ABT_VOCAB
        else:
            self.sense = {}
            add_vocab(vocab = VOCAB, name= self.name, sense_id= self.sense_id, sense= {})
            self.v = VOCAB
            logger.debug("Adding %s No registered sense, please initialize!", self.name)

        # don't know its rel yet

        for rel in self.v[self.name][sense_id].keys():
            # set them to implicit-argument
            self[(rel, {"implicit"})] = IMPLICIT

        self.sub_vars = {}

    def labels(self):
        raise NotImplementedError

# ********************************* relational operations **************************

    def __getitem__(self, rel_str):

        if rel_str == "ATTRIBUTES":
            rs = []
            for r in self.relations.keys():
                rs += self.relations[r]

            rel2attr = {}

            for r in rs:
                if isinstance(r.argument, Attribute) and (r.tags <= ALLOWED_TAGS):
                    rel2attr[r.name] = str(r.argument)

            # allow possible duplicated attributes

            return rel2attr

        elif rel_str == "ALL":
            rs = []
            for r in self.relations.keys():
                rs += self.relations[r]
        else:
            rs = self.relations[rel_str]

        nodes = []

        for r in rs:
            if r.tags <= ALLOWED_TAGS:
                nodes.append(r[self])

        return nodes

    def __setitem__(self, rel_and_tag, node):

        # r will be saved in reverse relation, but when output will be transformed

        rel, tags = rel_and_tag
        r = Relation(
            raw_name = rel,
            from_node= self,
            to_node = node,
            tags = tags
        ) # alias = )

        self.relations[rel].append(r)

        if type(node) == Word:
            node.relations[reverse(rel)].append(r)

    def __delitem__(self, rel):
        del self.relations[rel]

    def set_multiple_node_for_one_relation(self, rel: str, nodes: list):
        for node, is_down_direction in nodes:
            self[(rel, {"explicit"})] = node

    def __repr__(self):
        return self.name

# ********************************* this place above *******************************************************

    def to_str(self, sent_var):

        if self.sense_id > 0: #(len(self.v[self.name].keys())>1) or :
            return "(" + sent_var + self.var + " / " + self.name + "-{:02d}".format(self.sense_id)
        else:
            return "(" + sent_var + self.var + " / " + self.name

    def generate_umr(self, level, sent_var, var_list, rel_set, follow_old = False):

        s = []
        s.append(self.to_str(sent_var))   # "(" + self.var_name + " / " + self.name+"-0"+str(self.sense_id))
        for rel, rs in self.relations.items():

            for r in rs:

                node, is_down_direction, parapredicate = r[self]

                if follow_old and (not is_down_direction):
                    continue

                if not (r.tags <= ALLOWED_TAGS):
                    continue

                if type(node) == Attribute:
                    s.append(" "*4 * (level + 1)+":"+rel + " "*4 + node.generate_umr())
                    continue

                r_solid = r.solidate() # the two directional relations
                if r_solid in rel_set:
                    continue
                else:
                    rel_set.add(r_solid)

                if node.var not in var_list:
                    var_list.add(node.var)
                    s.append(" "*4 * (level + 1)+":"+rel + " "*4 + node.generate_umr(level + 1, sent_var, var_list, rel_set))
                    # var_list.pop(), don't pop, as all previous variable should be referred.
                else:
                    s.append(" "*4 * (level + 1)+":"+rel + " "*4 + sent_var + node.var)

        s[-1] += ")"

        return "\n".join(s)

    def get_relations(self):

        rel_set = set()

        for rs in self.relations.values():
            for r in rs:
                if r.tags <= ALLOWED_TAGS:
                    rel_set.add(r.solidate())

        return rel_set


class Relation:

    # registered = []

    def __init__(self, raw_name, from_node, to_node, tags, alias = None):

        """
        Things to do when initilizing an instance.
        """

        name = raw_name.lower()
        self.raw_name = raw_name
        self.name = reverse(name) if is_reverse(name) else name  # 也就是说name肯定是predicate 到 argument的
        self.tags = tags

        self.role = alias
        self.vector = None
        self.focus = from_node
        self.child = to_node

        self.parapredicate = False if is_reverse(name) else True # 顺磁性，意思是pred-argu关系和focus-child关系一致

        if self.parapredicate:
            self.predicate, self.argument = self.focus, self.child
        else:
            self.predicate, self.argument = self.child, self.focus

        if (type(to_node)==Word) and (type(from_node) == Word):
            self.between_var = True
        else:
            self.between_var = False

    def history(self):
        pass

    def __getitem__(self, node):
        """
        >>> node = Word()
        >>> self[node] = (another_node, "whether another_node is a child", "whether another_node is an argument")

        ______|_A->B__|__B-of->A__|
        r[A]  |       |           |
        ______|_B_T_T_|__B_F_T____|
        r[B]  |       |           |
        ______|_A_F_F_|__A_T_F____|


        """

        if node == self.focus:
            return self.child, True, self.parapredicate ^ False
        elif node == self.child:
            return self.focus, False, self.parapredicate ^ True


    def solidate(self):

        head = self.predicate
        tail = self.argument
        rel = self.name

        if self.between_var:
            return head.var, rel, tail.var
        else:
            if rel != "mod-of":
                try:
                    return head.var, rel, str(tail)
                except AttributeError as e:
                    logger.warning(f"Illegal head {head}, treated as string.")
            else:
                try:
                    return str(head), rel, tail.var
                except AttributeError as e:
                    logger.warning(f"Illegal node {tail}, treated as string.")



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Global default hyperparams"""

CNEIGHBOR: int = 1
SENSE_COEFFICIENT: float = 0.1
ALLOW_REIFY: bool = False
SEPARATE_1_AND_2: bool = False
USE_ALIGNMENT: bool = False
USE_SMATCH_TOP: bool = True
HANDLE_QUOTED_REENTRANCY = False



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Metric representation"""

from ancast.ops import protected_divide

class Metric:

    def __init__(self, name):
        self.name = name
        self.metrics = {
            "concept": {
            "good_sum": 0,
            "bad_quality_sum":0,
            "good_sum_count":0,
            "bad_quality_sum_count":0,
            "smatch_concept_sum": 0,
            "smatch_concept_sum_count":0
            },
            "relation": {
                "total": 0,
                "matched": 0
            },
            "lr":{
                "score": 0,
                "score_count": 0
            },
            "ulr":{
                "score": 0,
                "score_count": 0
            },
            "wlr":{
                "score": 0,
                "score_count": 0
            },
            "modal":{
                "score": 0,
                "score_count": 0
            },
            "temporal":{
                "score": 0,
                "score_count": 0
            },
            "coref":{
                "score": 0,
                "score_count": 0
            }
        }

    def log_and_inc_metric(self, Mt):
        for key1 in self.metrics:
            for key2 in self.metrics[key1]:
                self.metrics[key1][key2] += Mt.metrics[key1][key2]

    def log_and_inc(self, key1, key2, value, weight=1):
        assert key1 in self.metrics
        assert key2 in self.metrics[key1]
        self.metrics[key1][key2] += value
        self.metrics[key1][key2+"_count"] += weight

    def assign(self, key1, key2, value):
        self.metrics[key1][key2] = value

    def compute(self, type):
        if type == "smatch":
            return protected_divide(self.metrics["concept"]["smatch_concept_sum"] + self.metrics["relation"]["matched"] , self.metrics["concept"]["smatch_concept_sum_count"] + self.metrics["relation"]["total"])
        elif type == "concept with bad quality":
            return protected_divide((self.metrics["concept"]["good_sum"] + self.metrics["concept"]["bad_quality_sum"]) , (self.metrics["concept"]["good_sum_count"] + self.metrics["concept"]["bad_quality_sum_count"]))
        elif type == "concept qualified only":
            return protected_divide(self.metrics["concept"]["good_sum"] , self.metrics["concept"]["good_sum_count"] + self.metrics["concept"]["bad_quality_sum_count"])
            # unmatched is regarded as 0
        elif type == "relation":
            return protected_divide(self.metrics["relation"]["matched"], self.metrics["relation"]["total"])
        elif type == "good quality":
            return protected_divide(self.metrics["concept"]["good_sum_count"] , (self.metrics["concept"]["good_sum_count"] + self.metrics["concept"]["bad_quality_sum_count"]))
        elif type == "ulr":
            if self.metrics["ulr"]["score_count"] == 0:
                return self.compute("concept with bad quality")
            else:
                return protected_divide(self.metrics["ulr"]["score"], self.metrics["ulr"]["score_count"])
        elif type == "lr":
            if self.metrics["lr"]["score_count"] == 0:
                return self.compute("concept with bad quality")
            else:
                return self.metrics["lr"]["score"] / self.metrics["lr"]["score_count"]
        elif type == "wlr":
            if self.metrics["wlr"]["score_count"] == 0:
                return self.compute("concept with bad quality")
            else:
                return protected_divide(self.metrics["wlr"]["score"], self.metrics["wlr"]["score_count"])
        elif type == "modal":
            return protected_divide(self.metrics["modal"]["score"], self.metrics["modal"]["score_count"])
        elif type == "temporal":
            return protected_divide(self.metrics["temporal"]["score"], self.metrics["temporal"]["score_count"])
        elif type == "coref":
            return protected_divide(self.metrics["coref"]["score"], self.metrics["coref"]["score_count"])
        elif type == "document":
            return protected_divide(self.metrics["lr"]["score"] + self.metrics["modal"]["score"] + self.metrics["temporal"]["score"] + self.metrics["coref"]["score"], self.metrics["lr"]["score_count"]  + self.metrics["modal"]["score_count"] + self.metrics["temporal"]["score_count"] + self.metrics["coref"]["score_count"])
        else:
            return None



#! /usr/bin/python3
# -*- coding: utf-8 -*-
# Author: chunjy92
# Date: 12/4/24 17:41
from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


def main():
  pass


if __name__ == '__main__':
  main()



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Document representation"""

import csv
import logging
import re
from collections import defaultdict

from ancast import params
from ancast.closure import get_cluster
from ancast.matches import MatchResolution, Match
from ancast.metric import Metric
from ancast.ops import protected_divide, parse_alignment
from ancast.sentence import Sentence

logger = logging.getLogger(__name__)

SUBPATTERN = r'\(((\S*)\s*(\:\S*)\s*(\S*?))\)'


def get_edge_list_temp(tuple_dict):
    edge_list = defaultdict(list)
    # this will sometimes be problematic if debugging is calling "graph", this defaultdict
    for key, value in tuple_dict.items():
        if value == "after":
            edge_list[key[0]].append((key[1], "r"))
        elif value == "before":
            edge_list[key[1]].append((key[0], "r"))
        elif value == "depends-on":
            edge_list[key[0]].append((key[1], "dp"))
        elif value in {"contained", "includes"}:
            edge_list[key[1]].append((key[0], "dn"))
            edge_list[key[0]].append((key[1], "up"))
        elif value == "overlap":   # overlap sorted by number size
            min_key = min(key[0], key[1])
            max_key = max(key[0], key[1])
            edge_list[min_key].append((max_key, "o"))
        else:
            logger.info(f"{value} is an unrecognized relation in temporal!")

    return edge_list

def get_edge_list_coref(tuple_dict):
    edge_list = defaultdict(list)
    for key, value in tuple_dict.items():
        if value == "same-entity":
            edge_list[key[0]].append((key[1], "sn"))
            edge_list[key[1]].append((key[0], "sn"))
        elif value == "same-event":
            edge_list[key[1]].append((key[0], "sv"))
            edge_list[key[1]].append((key[0], "sv"))
        elif value == "subset-of":
            edge_list[key[0]].append((key[1], "up"))
            edge_list[key[1]].append((key[0], "dn"))
        else:
            logger.info("Unrecognized relation in coref!")

    return edge_list

def add_var(var_name, my_dict, other_dict, translate, match_list, n):
    if var_name not in my_dict:
        other = translate(match_list, var_name)
        if var_name in match_list.keys():
            my_dict[var_name] = n
            my_dict[n] = var_name
            if other != "NULL":
                other_dict[other] = n
                other_dict[n] = other
        else:
            # var_name is not a var, but const like author, root ...
            if var_name != other:
                logger.info(f"Translation error: inconsistant placeholder name {var_name} and {other} between test and gold!")
            my_dict[var_name] = n
            my_dict[n] = var_name
            other_dict[other] = n
            other_dict[n] = other
        n += 1

    return n


class SentenceMatch:

    # noinspection PyPep8Naming
    def __init__(
            self,
            format: str = 'umr',
            Cneighbor: int = params.CNEIGHBOR,
            sense_coefficient: float = params.SENSE_COEFFICIENT,
            allow_reify: bool = params.ALLOW_REIFY,
            separate_1_and_2: bool = params.SEPARATE_1_AND_2,
            use_alignment: bool = params.USE_ALIGNMENT,
            use_smatch_top: bool = params.USE_SMATCH_TOP,
    ):
        self.format = format
        self.Cneighbor = Cneighbor
        self.sense_coefficient = sense_coefficient
        self.allow_reify = allow_reify
        self.separate_1_and_2 = separate_1_and_2
        self.use_alignment = use_alignment
        self.use_smatch_top = use_smatch_top

        self.sentences = []
        self.smatch_scores = []
        self.concept_scores = []
        self.labeled_scores = []
        self.unlabeled_scores = []
        self.weighted_labeled_scores = []

        self.semantic_metric_precision = Metric("precision")
        self.semantic_metric_recall = Metric("recall")

        self.var_n = 0

        ### number of text data
        self.num_snts = 0
        self.num_toks = 0

        # final score
        self.sent_fscore = -1

        # per-anno messages
        self.msgs = []

    def macro_avg(self, M):
        self.smatch_scores.append(M.smatch_format_score)
        self.concept_scores.append(M.concept_match_fscore)
        self.labeled_scores.append(M.lbd_fscore)
        self.unlabeled_scores.append(M.ulbd_fscore)
        self.weighted_labeled_scores.append(M.wlbd_fscore)

    def add_doct_info(self, M, **kwargs):
        self.semantic_metric_precision.log_and_inc_metric(M.Mt01)
        self.semantic_metric_recall.log_and_inc_metric(M.Mt10)

    # def output_to_csv(self, writer, M = None, title=False):
    def summarize(self, match_res):
        formatted_str_match01 = '\n'.join(f"{Match.gname(match_res.umr0, key)}: {Match.gname(match_res.umr1, value)} \t Q-level = {match_res.quality_list01[key]}" for key, value in match_res.match_list01.items()) + "\n\nGood quality percentage:\t{:.2%}".format(match_res.gp01)
        formatted_str_match10 = '\n'.join(f"{Match.gname(match_res.umr1, key)}: {Match.gname(match_res.umr0, value)} \t Q-level = {match_res.quality_list10[key]}" for key, value in match_res.match_list10.items()) + "\n\nGood quality percentage:\t" + "{:.2%}".format(match_res.gp10)
        self.msgs.append(
            [
                match_res.umr0.text,
                match_res.umr0.generate_umr_text(),
                match_res.umr1.generate_umr_text(),
                formatted_str_match01,
                formatted_str_match10,
                "{:.2%}".format(match_res.match_score01),
                "{:.2%}".format(match_res.match_score10),
                "{:.2%}".format(match_res.concept_match_fscore),
                "{:.2%}".format(match_res.lbd_p),
                "{:.2%}".format(match_res.lbd_r),
                "{:.2%}".format(match_res.lbd_fscore),
                "{:.2%}".format(match_res.ulbd_fscore),
                "{:.2%}".format(match_res.wlbd_fscore),
                "{:.2%}".format(match_res.smatch_format_score)
            ]
        )

    def compute_scores(self, pred_inputs, gold_inputs):
        name, bgi = 0, -1
        for pred_input in pred_inputs:
            bgi += 1
            if "#" not in pred_input:
                logger.warning(f"Encountered unknown block in %s, skipping", bgi)
                continue

            pred_input_match = re.search(self.snt_pat, pred_input)
            if pred_input_match:  # the block is crucial
                while not re.search(self.snt_pat, gold_inputs[bgi]):
                    bgi += 1
                gold_input = gold_inputs[bgi]
                gold_input_match = re.search(self.snt_pat, gold_input)

                name += 1
                snt_test = pred_input.split(pred_input_match.group())[1].split("\n")[0].strip()
                snt_gold = gold_input.split(gold_input_match.group())[1].split("\n")[0].strip()
                # if snt_test != snt_gold:
                #     logger.info(f"{name} sentence is not the same!")

                if self.format_is_amr:
                    amr_test = "\n".join(pred_input.split("#")[-1].split("\n")[1:])
                    amr_gold = "\n".join(gold_input.split("#")[-1].split("\n")[1:])
                else:
                    amr_test = "\n".join(pred_input.split("\n\n")[1].split("\n")[1:])
                    amr_gold = "\n".join(gold_input.split("\n\n")[1].split("\n")[1:])
                ta, ga = {}, {}

                tamr = Sentence(
                    sent=snt_test,
                    semantic_text=amr_test,
                    alignment=ta,
                    sent_num=name,
                    format=self.format
                )
                gamr = Sentence(
                    sent=snt_gold,
                    semantic_text=amr_gold,
                    alignment=ga,
                    sent_num=name,
                    format=self.format
                )

                if tamr.invalid or gamr.invalid:
                    logger.warning(f"Error encountered, skipping sentence {name}")
                    continue

                match_res = MatchResolution(
                    tamr,
                    gamr,
                    Cneighbor=self.Cneighbor,
                    sense_coefficient=self.sense_coefficient,
                    separate_1_and_2=self.separate_1_and_2,
                    use_alignment=self.use_alignment,
                    use_smatch_top=self.use_smatch_top,
                )

                self.macro_avg(match_res)
                self.add_doct_info(match_res)  # micro-average

                # cache msgs for analysis
                self.summarize(match_res)

        # aggregate
        ps, rs = self.semantic_metric_precision.compute("lr"), self.semantic_metric_recall.compute("lr")
        self.sent_fscore = protected_divide(2*ps*rs, ps+rs)
        logger.info(
            f"Sent Micro:\tPrecision: {ps:.2%}\tRecall: {rs:.2%}\tFscore: {self.sent_fscore:.2%}"
        )

        return self.sent_fscore

    @property
    def format_is_amr(self):
        return  self.format == "amr"

    @property
    def snt_pat(self):
        return  r"::snt\s" if self.format_is_amr else r"::\ssnt\d+"


class DocumentMatch(SentenceMatch):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.doc_annotations_test = {"temporal":  {}, "modal": {}, "coref": {}}
        self.doc_annotations_gold = {"temporal":  {}, "modal": {}, "coref": {}}
        self.doc_var_list = {"test" : {}, "gold":{}}

        self.doc_num_snts = -1
        self.doc_num_toks = -1

        # final scores
        self.modal_fscore = -1
        self.temporal_fscore = -1
        self.coref_fscore = -1
        self.comp_fscore = -1

        # per-anno messages
        self.msgs = []

    def add_doct_info(self, M, **kwargs):
        super().add_doct_info(M, **kwargs)

        test_doc = kwargs["test_doc"]
        gold_doc = kwargs["gold_doc"]

        patterns = {
            'temporal': r':temporal\s*\(((\(.*\:.*\)\s*)*)\)',
            'modal': r':modal\s*\(((\(.*\:.*\)\s*)*)\)',
            'coref': r':coref\s*\(((\(.*\:.*\)\s*)*)\)'
        }

        for key, pattern in patterns.items():
                match = re.search(pattern, test_doc)
                if match:
                    ms = match.group(1)
                    rs = re.findall(SUBPATTERN, ms)
                    for mts in rs:
                        head = mts[1]
                        rel = mts[2]
                        tail = mts[3]
                        self.var_n = add_var(head, self.doc_var_list["test"], self.doc_var_list["gold"], M.translate_match, M.match_list01, self.var_n)
                        self.var_n = add_var(tail, self.doc_var_list["test"], self.doc_var_list["gold"], M.translate_match, M.match_list01, self.var_n)
                        self.doc_annotations_test[key][(self.doc_var_list["test"][head], self.doc_var_list["test"][tail])] = rel[1:]

                match = re.search(pattern, gold_doc)
                if match:
                    ms = match.group(1)
                    rs = re.findall(SUBPATTERN, ms)
                    for mts in rs:
                        head = mts[1]
                        rel = mts[2]
                        tail = mts[3]
                        self.var_n = add_var(head, self.doc_var_list["gold"], self.doc_var_list["test"], M.translate_match, M.match_list10, self.var_n)
                        self.var_n = add_var(tail, self.doc_var_list["gold"], self.doc_var_list["test"], M.translate_match, M.match_list10, self.var_n)
                        self.doc_annotations_gold[key][(self.doc_var_list["gold"][head], self.doc_var_list["gold"][tail])] = rel[1:]

    def compute_scores(self, pred_inputs, gold_inputs):
        name = num_snts = num_toks = 0

        # ignoring file head
        for pred_input, gold_input in zip(pred_inputs, gold_inputs):
            name += 1
            logger.debug("Sentence %d"%name)

            try:
                assert "sentence" in pred_input and "sentence" in gold_input, \
                    f"Keyword `sentence` is not found in block {name}"
            except AssertionError as error:
                logger.info(f"Format Error: {error.args[0]}")
                raise
            try:
                assert "document" in pred_input and "document" in gold_input, \
                    f"Keyword `document` is not found in block {name}"
            except AssertionError as error:
                logger.info(f"Format Error: {error.args[0]}")
                raise

            t_buff = pred_input.split("# sentence level graph:")
            g_buff = gold_input.split("# sentence level graph:")

            t_sent = re.sub(r'^\d+[\s\t]*', '', t_buff[0]).strip()
            g_sent = re.sub(r'^\d+[\s\t]*', '', g_buff[0]).strip()

            t_buff = t_buff[1].strip().split("# alignment:")
            g_buff = g_buff[1].strip().split("# alignment:")

            t_graph = t_buff[0].strip()
            g_graph = g_buff[0].strip()

            t_buff = t_buff[1].strip().split("# document level annotation:")
            g_buff = g_buff[1].strip().split("# document level annotation:")

            t_alignment = t_buff[0].strip()
            g_alignment = g_buff[0].strip()

            t_doclevel = t_buff[1].strip()
            g_doclevel = g_buff[1].strip()

            t_alignment = parse_alignment(t_alignment)
            g_alignment = parse_alignment(g_alignment)

            if t_sent != g_sent:
                logger.info(f"Warning: Sentence {name} sentence content mismatched! Using test sentence.")
                g_sent = t_sent

            tumr = Sentence(
                sent = t_sent,
                semantic_text = t_graph,
                alignment= t_alignment,
                sent_num = name,
                format = self.format
            )
            gumr = Sentence(
                sent = g_sent,
                semantic_text = g_graph,
                alignment= g_alignment,
                sent_num = name,
                format = self.format
            )

            if tumr.invalid or gumr.invalid:
                logger.info(f"Error encountered, skipping sentence {name}")
                continue

            match_res = MatchResolution(
                tumr,
                gumr,
                Cneighbor=self.Cneighbor,
                sense_coefficient=self.sense_coefficient,
                separate_1_and_2=self.separate_1_and_2,
                use_alignment=self.use_alignment,
                use_smatch_top=self.use_smatch_top,
            )

            self.add_doct_info(match_res, test_doc = t_doclevel, gold_doc = g_doclevel)
            self.macro_avg(match_res)

            self.summarize(match_res)

            num_snts += 1
            if t_sent is not None:
                num_toks += len(t_sent.split())

        self.doc_num_snts = num_snts
        self.doc_num_toks = num_toks

        ps, rs = self.semantic_metric_precision.compute("lr"), self.semantic_metric_recall.compute("lr")
        self.sent_fscore = protected_divide(2*ps*rs, ps+rs)
        logger.info(f"Sent Micro:\tPrecision: {ps:.2%}\tRecall: {rs:.2%}\tFscore: {self.sent_fscore:.2%}")

        pm, rm, self.modal_fscore = self.calculate_modal()
        pt, rt, self.temporal_fscore = self.calculate_TCTC("temporal", get_edge_list_temp)
        pc, rc, self.coref_fscore = self.calculate_TCTC("coref", get_edge_list_coref)

        logger.info(f"Modality:\tPrecision: {pm:.2%}\tRecall: {rm:.2%}\tFscore: {self.modal_fscore:.2%}")
        logger.info(f"Temporal:\tPrecision: {pt:.2%}\tRecall: {rt:.2%}\tFscore: {self.temporal_fscore:.2%}")
        logger.info(f"Coref:\t\tPrecision: {pc:.2%}\tRecall: {rc:.2%}\tFscore: {self.coref_fscore:.2%}")

        comp_p = self.semantic_metric_precision.compute("document")
        comp_r = self.semantic_metric_recall.compute("document")

        self.comp_fscore = protected_divide(2*comp_p*comp_r, comp_p+comp_r)

        logger.info(f"Comprehensive Score:\t{self.comp_fscore:.2%}\n")

        return self.sent_fscore, self.modal_fscore, self.temporal_fscore, self.coref_fscore, self.comp_fscore


    def warn_error(self, theset, test_or_gold, temp_or_coref):
        if theset:
            elem = next(iter(theset))
            if len(elem) == 1:
                logger.info(f"These nodes have erratic circular relations in {temp_or_coref} in {test_or_gold} file.")
                pstr = " ".join([self.doc_var_list[test_or_gold][n[0]] for n in theset])
            elif len(elem) == 2:
                logger.info(f"These pairs of nodes have paradoxical relations in {temp_or_coref} in {test_or_gold} file.")
                pstr = "\t".join(["("+self.doc_var_list[test_or_gold][n[0]]+", "+self.doc_var_list[test_or_gold][n[1]]+")" for n in theset])
            elif len(elem) == 3:
                logger.info(f"These nodes triples have contradictory coreferential relations in {test_or_gold}.")
                pstr = "\t".join(["(" + ", ".join([self.doc_var_list[test_or_gold][n[i]] for i in range(3)])  +")" for n in theset])

            logger.info(pstr)

    def calculate_modal(self):
        set0, set1 = set(), set()
        for key, value in self.doc_annotations_test["modal"].items():
            set0.add(str(key)+str(value))
        for key,value in self.doc_annotations_gold["modal"].items():
            set1.add(str(key)+str(value))

        set2 = set0 & set1
        modal_precision = protected_divide(len(set2) , len(set0))
        modal_recall    = protected_divide(len(set2) , len(set1))

        self.semantic_metric_precision.log_and_inc("modal", "score",len(set2), len(set0))
        self.semantic_metric_recall.log_and_inc("modal", "score", len(set2), len(set1))

        return modal_precision, modal_recall, protected_divide(2*modal_precision*modal_recall , (modal_precision + modal_recall))

    def calculate_TCTC(self, keyword, get_f):
        """
        1. Organize temporal relation all into uni-direction
        2. Make this uni-directed graph a edge list;
        3. Get this edge list into Closure and use DFS to get a closure
        4. Use the result of two closure to calculate

        depends-on, overlap
        contain (only going down), after, before
        """

        el_test =  get_f(self.doc_annotations_test[keyword])
        el_gold =  get_f(self.doc_annotations_gold[keyword])

        # two steps above are pure translation
        cluster_el_test, test_whole_set, inconsistent_test, circular_test, wrongcorefs_test = get_cluster(el_test)
        cluster_el_gold, gold_whole_set, inconsistent_gold, circular_gold, wrongcorefs_gold = get_cluster(el_gold)

        self.warn_error(inconsistent_test, "test", keyword)
        self.warn_error(inconsistent_gold, "gold", keyword)
        self.warn_error(circular_test, "test", keyword)
        self.warn_error(circular_gold, "gold", keyword)
        self.warn_error(wrongcorefs_test, "test", keyword)
        self.warn_error(wrongcorefs_gold, "gold", keyword)

        s0 = 0
        cnt_test = 0
        for cluster, associated_nodes in cluster_el_test:
            cluster_size = len(associated_nodes)
            s0 += cluster_size *  protected_divide(len(cluster & gold_whole_set) , len(cluster))
            # here we need to be careful that there may be some ill-formed standing-alone pairs
            cnt_test += cluster_size

        self.semantic_metric_precision.log_and_inc(keyword, "score", s0, cnt_test)
        s0 = protected_divide(s0, cnt_test)

        s1 = 0
        cnt_gold = 0
        for cluster, associated_nodes in cluster_el_gold:
            cluster_size = len(associated_nodes)
            s1 += cluster_size *  len(cluster & test_whole_set) / len(cluster)
            cnt_gold += cluster_size
        self.semantic_metric_recall.log_and_inc(keyword, "score", s1, cnt_gold)
        s1 = protected_divide(s1, cnt_gold)

        return s0, s1, protected_divide(2*s0*s1 , (s0+s1))



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Match representation"""

import logging
from collections import defaultdict

import numpy as np

from ancast import ops
from ancast.metric import Metric
from ancast.word import Word

logger = logging.getLogger(__name__)


class Match:

    def __init__(self, umr0, umr1):
        self.umr0 = umr0
        self.umr1 = umr1
        self.var2idx = None


    def draw_matrix(self, matrices):
        import matplotlib.pyplot as plt

        row_names = [self.umr0.var2node[self.var2idx[0][i]] for i in range(matrices[0].shape[0])]
        col_names = [self.umr1.var2node[self.var2idx[1][i]] for i in range(matrices[0].shape[1])]


        ncols = (len(matrices) + 1) // 2

        fig, axes = plt.subplots(nrows=2, ncols=ncols, figsize=(15, 8))

        for i, matrix in enumerate(matrices):
            im = axes[(i) // ncols, (i) % ncols].imshow(matrix, cmap='viridis', interpolation='none', vmin=0, vmax=1)
            axes[(i) // ncols, (i) % ncols].set_title(f"Iteration {i+1}")
            axes[(i) // ncols, (i) % ncols].set_xticks(range(len(col_names)))
            axes[(i) // ncols, (i) % ncols].set_xticklabels(col_names, rotation=90)
            axes[(i) // ncols, (i) % ncols].set_yticks(range(len(row_names)))
            axes[(i) // ncols, (i) % ncols].set_yticklabels(row_names)

        cbar_ax = fig.add_axes([0.92, 0.15, 0.02, 0.7])
        fig.colorbar(im, cax=cbar_ax)

        plt.tight_layout()
        plt.subplots_adjust(right=0.9)
        plt.show()

    def translate_match(self, match_list, var):
        if var in match_list.keys():
            return match_list[var]
        elif var == "TOP":
            return "TOP"
        else:
            return var # if it is an Attribute, all attributes should be stored in lower case

    def structural_count(self, use_smatch_top=False):

        vlist0 = self.umr0.var2node
        vlist1 = self.umr1.var2node

        set0 = set()
        set1 = set()

        for node in vlist0.values():
            set0 = set0 | node.get_relations()
        for node in vlist1.values():
            set1 = set1 | node.get_relations()

        if use_smatch_top:
            set0.add(("TOP", "TOP", self.var2idx[0][0]))
            set1.add(("TOP", "TOP", self.var2idx[1][0]))

        # translate 0-1 and compare overlap, all vars in 0 are guaranteed to be translated
        translated_set0 = set()

        for r in set0:
            translated_set0.add(
                (self.translate_match(self.match_list01, r[0]), r[1], self.translate_match(self.match_list01,r[2]))
            )

        translated_set1 = set()
        for r in set1:
            translated_set1.add(
                (self.translate_match(self.match_list10, r[0]), r[1], self.translate_match(self.match_list10, r[2]))
            )

        return len(translated_set1 & set0), len(set0), len(translated_set0 & set1), len(set1)

    @staticmethod
    def gname(umr, v):
        if v in umr.var2node.keys():
            return f"{v} / {str(umr.var2node[v])}"
        else:
            try:
                assert v[:4] == "NULL", "Unmatched nodes not mapped to NULL!"
            except AssertionError as e:
                print(e.args[0])
                raise
            return "null"

    @staticmethod
    def transform_alignment(umr):
        for x in umr.var2node.keys():
            if x in umr.alignment:
                pair_text = umr.alignment[x]
                pair = [0, 0]
                if (pair_text[0] == "0") or (pair_text[0] == "-"):
                    pair = [0, 0]
                else:
                    pair = pair_text.split("-")
                    if "undefined" in pair:
                        pair = [0, 0]
                    pair[0] = int(pair[0])
                    pair[1] = int(pair[1])
                umr.alignment[x] = pair
            else:
                umr.alignment[x] = [0,0]

    @staticmethod
    def transform_mapping(match_list, var2node):
        t = {}
        for key in match_list.keys():
            if match_list[key] != "null":
                t[var2node[0][key]] = var2node[1][match_list[key]]
            else:
                t[var2node[0][key]] = "null"
        return t

    @staticmethod
    def enhance_with_rel_label(idx, maxima, umr_idx,  var2idx, umr0, umr1):

        """
        Used to select the most likely pairs based on relation label similarity, when they have the same non-relational similarity.
        """

        max_sim = -1
        arg_max_sim = None
        label_sim = np.zeros(maxima.shape)
        umr_this = umr0 if umr_idx == 0 else umr1
        umr_that = umr1 if umr_idx == 0 else umr0

        rel_set_this = Match.get_rel_set(idx, var2idx[umr_idx], umr_this.var2node)
        for k, ii in enumerate(maxima):
            rel_set_that = Match.get_rel_set(ii, var2idx[1 - umr_idx], umr_that.var2node)
            temp_s = ops.jaccard_sim(rel_set_this, rel_set_that)
            label_sim[k] = temp_s

        max_label_sim = np.amax(label_sim)
        if np.count_nonzero(label_sim == max_label_sim) > 1:
            return None, np.where(label_sim == max_label_sim)[0]
        else:
            return maxima[np.argmax(label_sim)], np.where(label_sim == max_label_sim)[0]

    @staticmethod
    def get_rel_set(idx, var2idx, var2node):
        return set(var2node[var2idx[idx]].relations.keys())


class MatchResolution(Match):

    def __init__(
            self,
            umr0,
            umr1,
            Cneighbor: int = 1,
            sense_coefficient: float = 0.1,
            separate_1_and_2: bool = False,
            use_alignment: bool = False,
            use_smatch_top: bool = True,
    ):
        super().__init__(umr0, umr1)

        # get name matrix for evaluation and anchors for iterative neighbor information
        self.Mt01 = Metric("01")
        self.Mt10 = Metric("10")

        self.initial_sim, self.var2idx, self.anchors, self.name_idt = self.get_global_match(
            self.umr0, self.umr1, anchor_with_alignment=use_alignment, sense_coefficient=sense_coefficient
        )

        self.N_Matrices, self.outmatrix = self.get_neighbor_matrix(separate_1_and_2=separate_1_and_2)

        self.match_list01, self.match_list10, self.quality_list01, self.quality_list10 = self.iterate_anchor_and_match(
            self.anchors, add_asymmetric_relations=True, Cneighbor=Cneighbor, separate_1_and_2=separate_1_and_2,
        )

        _, _, _, self.smatch_format_score = self.get_structural_plain_count(use_smatch_top)

        self.lbd_p, self.lbd_r, self.lbd_fscore, self.ulbd_fscore, self.wlbd_fscore = self.weighted_relational_overlap(
            sense_coefficient=sense_coefficient,
        )

    def get_structural_plain_count(self, use_smatch_top=False):

        overlap0, all0, overlap1, all1 = self.structural_count(use_smatch_top=use_smatch_top)

        self.Mt01.assign("relation", "matched", overlap0)
        self.Mt01.assign("relation", "total",   all0)
        self.Mt10.assign("relation", "matched", overlap1)
        self.Mt10.assign("relation", "total",   all1)

        self.gp01 = self.Mt01.compute("good quality")
        self.gp10 = self.Mt10.compute("good quality")

        self.match_score01 = self.Mt01.compute("concept with bad quality")
        self.match_score10 = self.Mt01.compute("concept with bad quality")

        self.concept_match_fscore = ops.protected_divide((self.match_score01 * self.match_score10 * 2) , (self.match_score01 + self.match_score10))

        precision = self.Mt01.compute("relation")
        recall = self.Mt10.compute("relation")

        if precision + recall > 0:
            sm_fscore = 2 * precision * recall / (precision + recall)
        else:
            sm_fscore = 0.0

        psm = self.Mt01.compute("smatch")
        rsm = self.Mt10.compute("smatch")
        smatch_format_score = ops.protected_divide(psm * rsm * 2, (psm + rsm))

        return precision, recall, sm_fscore, smatch_format_score

    def get_global_match(
            self,
            umr0,
            umr1,
            anchor_with_alignment = False,
            sense_coefficient=0.1,
    ):

        vlist0 = umr0.var2node
        vlist1 = umr1.var2node

        Match.transform_alignment(umr0)
        Match.transform_alignment(umr1)

        anchors = np.zeros((len(vlist0), len(vlist1)))
        initial_sim = np.zeros((len(vlist0), len(vlist1)))
        name_idt = np.zeros((len(vlist0), len(vlist1)))

        var2idx = ({},{})

        name_duplicate_0 = defaultdict(lambda: 0)
        name_duplicate_1 = defaultdict(lambda: 0)

        for v0 in vlist0:
            name_duplicate_0[vlist0[v0].name] += 1
        for v1 in vlist1:
            name_duplicate_1[vlist1[v1].name] += 1

        for i, v0 in enumerate(vlist0):
            for j, v1 in enumerate(vlist1):

                var2idx[0][v0] = i
                var2idx[1][v1] = j
                var2idx[0][i] = v0
                var2idx[1][j] = v1

                node0 = umr0.var2node[v0]
                node1 = umr1.var2node[v1]

                # sense check

                if (node0.sense_id == node1.sense_id):
                    sense_sim = 1
                else:
                    sense_sim = 0

                # attribute check

                v0rel_attr = vlist0[v0]['ATTRIBUTES']
                v1rel_attr = vlist1[v1]['ATTRIBUTES']

                # "ops" would be combined into string, other attributes compared separately
                n = 0
                s = 0
                op0 = []
                for rel in v0rel_attr.keys():
                    if rel[:2] == "op":
                        op0.append(v0rel_attr[rel])
                        continue
                    if rel in v1rel_attr.keys():
                        n += 1
                        s += (v0rel_attr[rel] == v1rel_attr[rel])

                op1 = []
                for rel in v1rel_attr.keys():
                    if rel[:2] == "op":
                        op1.append(v1rel_attr[rel])

                op0 = "".join(op0)
                op1 = "".join(op1)

                # final initial similarity and name similarity
                this_name_sim = ops.name_sim(
                    node0.name+op0,
                    node1.name+op1,
                    sense_sim,
                    sense_coefficient=sense_coefficient
                )

                other_attr = s / n if n > 0 else 0
                ot_at_co = 1 if n > 0 else 0            # 不应该是attributes不一样就不算，而是没有的话就不算

                initial_sim[i, j] = (this_name_sim  + other_attr) / (1 + ot_at_co)
                name_idt[i, j] = (this_name_sim == 1) and (sense_sim == 1)
                # anchors

                # alignment
                if anchor_with_alignment:
                    min_low = min(umr0.alignment[v0][0], umr1.alignment[v1][0])
                    max_low = max(umr0.alignment[v0][0], umr1.alignment[v1][0])
                    min_high = min(umr0.alignment[v0][1], umr1.alignment[v1][1])
                    max_high = max(umr0.alignment[v0][1], umr1.alignment[v1][1])

                    maximum_overlap = min_high - max_low + 1
                    maximum_range = max_high - min_low + 1

                    if min_low == 0:
                        align_sim = 0
                    elif maximum_overlap <= 0:
                        align_sim = 1e-2
                    else:
                        align_sim = maximum_overlap / maximum_range # alignment可以用

                    if align_sim == 1:
                        anchors[i, j] = 1

                else:
                    # if both of them only appear once in the original sentence, then they are unique and anchored

                    if (node0.name == node1.name) and (node0.name in umr0.text.lower()) and (node1.name in umr1.text.lower()) and (name_duplicate_0[node0.name] == 1) and (name_duplicate_1[node1.name] == 1):
                        anchors[i, j] = 1

        return initial_sim, var2idx, anchors, name_idt

    # @timer_decorator
    def get_neighbor_matrix(
            self,
            predarguRatherthanfocuschild = True,
            two_layer_neighbor = True,
            separate_1_and_2=False,
    ):

        if predarguRatherthanfocuschild:
            node_idx = 2
        else:
            node_idx = 1

        vlist0 = self.umr0.var2node
        vlist1 = self.umr1.var2node

        Ain_umr0 = np.zeros((len(vlist0), len(vlist0)))
        Ain_umr1 = np.zeros((len(vlist1), len(vlist1)))

        Aout_umr0 = np.zeros((len(vlist0), len(vlist0)))
        Aout_umr1 = np.zeros((len(vlist1), len(vlist1)))

        for v0 in vlist0:
            v0idx = self.var2idx[0][v0]
            neighbors_v0 = vlist0[v0]['ALL']
            v0nb_idxes_out = [self.var2idx[0][node[0].var] for node in neighbors_v0 if (type(node[0]) == Word) and node[node_idx]]
            v0nb_idxes_in = [self.var2idx[0][node[0].var] for node in neighbors_v0 if (type(node[0]) == Word) and not node[node_idx]]

            Aout_umr0[v0idx][v0nb_idxes_out] = 1
            Ain_umr0[v0idx][v0nb_idxes_in]   = 1

        for v1 in vlist1.keys():
            v1idx = self.var2idx[1][v1]
            neighbors_v1 = vlist1[v1]['ALL']
            v1nb_idxes_out = [self.var2idx[1][node[0].var] for node in neighbors_v1 if (type(node[0]) == Word) and node[node_idx]]
            v1nb_idxes_in = [self.var2idx[1][node[0].var] for node in neighbors_v1 if (type(node[0]) == Word) and not node[node_idx]]
            Aout_umr1[v1idx][v1nb_idxes_out] = 1
            Ain_umr1[v1idx][v1nb_idxes_in]   = 1

        # adding grandparent, grandson, and sibling information
        # grandparent and grandson is square of matrix
        # in matrix A, out matrix B, then with grand informaion (A^2 + A) and (B^2 + B)
        # siblings are (AB + BA) - I (excluding itself)
        # matrix is very elegant, except that >1 need to be 1

        outmatrix = Aout_umr0, Aout_umr1

        if separate_1_and_2:
            M0 = np.ones(Aout_umr0.shape) - np.eye(len(vlist0)) # J - I
            M1 = np.ones(Aout_umr1.shape) - np.eye(len(vlist1))

            Aout_umr0_2hop = np.matmul(Aout_umr0, Aout_umr0) * M0
            Aout_umr1_2hop = np.matmul(Aout_umr1, Aout_umr1) * M1
            Ain_umr0_2hop  = np.matmul(Ain_umr0, Ain_umr0)   * M0
            Ain_umr1_2hop  = np.matmul(Ain_umr1, Ain_umr1)   * M1

            Aout_umr0_2hop[Aout_umr0 > 1] = 1
            Aout_umr1_2hop[Aout_umr1 > 1] = 1
            Ain_umr0_2hop[Ain_umr0 > 1] = 1
            Ain_umr1_2hop[Ain_umr1 > 1] = 1

            return (Aout_umr0, Aout_umr0_2hop, Aout_umr1, Aout_umr1_2hop, Ain_umr0, Ain_umr0_2hop, Ain_umr1, Ain_umr1_2hop, 1, 1), outmatrix

        else:
            M0 = np.ones(Aout_umr0.shape) - np.eye(len(vlist0)) # J - I
            M1 = np.ones(Aout_umr1.shape) - np.eye(len(vlist1))

            Aout_umr0 = (np.matmul(Aout_umr0, Aout_umr0) + Aout_umr0) * M0
            Aout_umr1 = (np.matmul(Aout_umr1, Aout_umr1) + Aout_umr1) * M1
            Ain_umr0  = (np.matmul(Ain_umr0, Ain_umr0)   +  Ain_umr0) * M0
            Ain_umr1  = (np.matmul(Ain_umr1, Ain_umr1)   +  Ain_umr1) * M1

            Aout_umr0[Aout_umr0 > 1] = 1
            Aout_umr1[Aout_umr1 > 1] = 1
            Ain_umr0[Ain_umr0 > 1] = 1
            Ain_umr1[Ain_umr1 > 1] = 1

            s_Aout_umr0 = Aout_umr0.sum(axis = 1)[:, np.newaxis] # how many 1 & 2-hop neighbors in total
            s_Aout_umr1 = Aout_umr1.sum(axis = 1)[np.newaxis, :]
            s_Ain_umr0 =  Ain_umr0.sum(axis = 1)[:, np.newaxis]
            s_Ain_umr1 =  Ain_umr1.sum(axis = 1)[np.newaxis, :]


            denom_out = np.minimum(s_Aout_umr0, s_Aout_umr1)   #  1 & 5 neighbors v. 2 & 2 neighbors, the former should be lower.
            denom_in  = np.minimum(s_Ain_umr0, s_Ain_umr1)
            denom_out[denom_out==0] = np.inf
            denom_in[denom_in==0]   = np.inf

            C_out = np.maximum(s_Aout_umr0, s_Aout_umr1) / denom_out
            C_in = np.maximum(s_Ain_umr0, s_Ain_umr1) / denom_in
            C_out[C_out==0] = np.inf
            C_in[C_in==0]  = np.inf

            return (Aout_umr0, Aout_umr1, Ain_umr0, Ain_umr1, C_out, C_in), outmatrix

    # @timer_decorator
    def incorporate_neighborhood_information(
            self,
            initial_sim,
            anchors,
            Cneighbor=1 ,
            max_iterations = 50,
            tolerance=1e-4,
            alpha = 2e-1,
            beta = 1e-3,
            separate_1_and_2=False,
    ):

        if separate_1_and_2:
            Aout_umr0_1hop, Aout_umr0_2hop, Aout_umr1_1hop, Aout_umr1_2hop, Ain_umr0_1hop, Ain_umr0_2hop, Ain_umr1_1hop, Ain_umr1_2hop, C_out, C_in = self.N_Matrices
        else:
            Aout_umr0, Aout_umr1, Ain_umr0, Ain_umr1, C_out, C_in = self.N_Matrices

        newsim = anchors.copy() # note that initial_sim_matrix is not necessarily a symmetric matrix!

        matrices = [newsim]

        for its in range(max_iterations):
            prevsim = newsim.copy()
            if separate_1_and_2:

                Sout = (Aout_umr0_1hop @ prevsim) @ Aout_umr1_1hop.T + (Aout_umr0_2hop @ prevsim) @ Aout_umr1_2hop.T
                Sin  = (Ain_umr0_1hop  @ prevsim) @  Ain_umr1_1hop.T + (Ain_umr0_2hop  @ prevsim) @  Ain_umr1_2hop.T

            else:

                Sout = (Aout_umr0 @ prevsim) @ Aout_umr1.T / C_out
                Sin  = (Ain_umr0  @ prevsim) @  Ain_umr1.T / C_in

            newsim = np.sqrt((Sout + 1) * (Sin + 1) - 1) # + Ssib

            newsim_max = np.amax(newsim)
            if newsim_max > 0:
                newsim = newsim / newsim_max

            newsim = newsim * Cneighbor

            newsim = np.maximum(newsim, anchors)

            if np.allclose(prevsim, newsim, atol=tolerance):
                break

            if its % 5 == 0:
                matrices.append(newsim)

        logger.debug(f"Converge in %d rounds.", its)

        decimal = int(round(-np.log10(tolerance)))
        return np.around((newsim +beta) * (initial_sim + alpha), decimals = decimal), newsim, matrices

    # @timer_decorator
    def greedy_match_list(self, final_sim, unmatched_list_01, unmatched_list_10, final_match_01, final_match_10, quality_list01, quality_list10):
        sub_matrix = final_sim[np.ix_(unmatched_list_01, unmatched_list_10)]

        order_max_to_min_01 = final_sim.argsort(axis=1)[:, ::-1]
        order_max_to_min_10 = final_sim.argsort(axis=0).T[:, ::-1]

        match_list_01 = {}
        match_list_10 = {}
        matched_place_01 = np.zeros(len(order_max_to_min_01), dtype=np.int64)
        matched_place_10 = np.zeros(len(order_max_to_min_10), dtype=np.int64)

        for i in unmatched_list_01:
            match_list_01[i] = order_max_to_min_01[i][matched_place_01[i]]
        for i in unmatched_list_10:
            match_list_10[i] = order_max_to_min_10[i][matched_place_10[i]]

        cnt = 0
        while match_list_01 and match_list_10:

            cnt += 1
            index_of_unmatch_01 = {unmatched_list_01[i]:i for i in range(len(unmatched_list_01))}
            index_of_unmatch_10 = {unmatched_list_10[j]:j for j in range(len(unmatched_list_10))}

            for i, idx in enumerate(match_list_01.keys()):

                match_list_01[idx] = order_max_to_min_01[idx][matched_place_01[idx]]

                while match_list_01[idx] not in unmatched_list_10:
                # while self.var2idx[1][match_list_01[idx]] in final_match_10.keys():
                    matched_place_01[idx] += 1
                    if matched_place_01[idx] < len(order_max_to_min_01[idx]):
                        match_list_01[idx] = order_max_to_min_01[idx][matched_place_01[idx]]
                    else:
                        break

                # if there is a tied maximum, go look at their relation labels

                try:
                    maxima = np.where((sub_matrix[i] == sub_matrix[i, index_of_unmatch_10[match_list_01[idx]]])>0)[0]
                except:
                    print(sub_matrix[i], "\n", match_list_01[idx], "\n", index_of_unmatch_10[match_list_01[idx]])

                if len(maxima) != 1:
                    vfunc = np.vectorize(lambda x: unmatched_list_10[x])
                    mapped_maxima = vfunc(maxima)
                    tgt, wheres = Match.enhance_with_rel_label(idx, mapped_maxima, 0, self.var2idx, self.umr0, self.umr1)
                    if tgt:
                        assert tgt in match_list_10.keys() # otherwise implementation error, tgt must be in mapped maxima
                    else:
                        # tied, possibly two relations are both 0
                        tgt = mapped_maxima[wheres[0]] # $wheres$ only select from those maximums
                    match_list_01[idx] = tgt
                    if cnt>500:
                        print(f"{i}-{idx}", wheres)
                    matched_place_01[idx] = max(matched_place_01[idx] - len(maxima), 0)
                    # it's possible that another choice is skipped

            for j, idx in enumerate(match_list_10.keys()):

                match_list_10[idx] = order_max_to_min_10[idx][matched_place_10[idx]]

                while match_list_10[idx] not in unmatched_list_01:
                # while self.var2idx[0][match_list_10[idx]] in final_match_01.keys():
                    matched_place_10[idx] += 1
                    if matched_place_10[idx] < len(order_max_to_min_10[idx]):
                        match_list_10[idx] = order_max_to_min_10[idx][matched_place_10[idx]]
                    else:
                        break

                maxima = np.where((sub_matrix[:, j] == sub_matrix[index_of_unmatch_01[match_list_10[idx]], j])>0)[0]

                if len(maxima) != 1:
                    vfunc = np.vectorize(lambda x: unmatched_list_01[x])
                    mapped_maxima = vfunc(maxima)
                    tgt, wheres = Match.enhance_with_rel_label(idx, mapped_maxima, 1, self.var2idx, self.umr0, self.umr1)
                    if tgt:
                        assert tgt in match_list_01.keys() # otherwise implementation error, tgt must be in mapped maxima
                    else:
                        tgt = mapped_maxima[wheres[0]]
                    if cnt>500:
                        print(f"{j}-||{idx}", wheres)
                    match_list_10[idx] = tgt
                    matched_place_10[idx] = max(matched_place_10[idx] - len(maxima), 0)
                    # it's possible that another choice is skipped

            key_pairs_to_delete = set()

            for idx01, idx10 in match_list_01.items():
                if (idx10 in match_list_10.keys()) and (match_list_10[idx10] == idx01):
                    assert self.var2idx[0][idx01] not in final_match_01.keys()
                    assert self.var2idx[1][idx10] not in final_match_10.keys()

                    final_match_01[self.var2idx[0][idx01]] = self.var2idx[1][idx10]
                    quality_list01[self.var2idx[0][idx01]] = -1
                    self.Mt01.log_and_inc("concept", "bad_quality_sum", self.initial_sim[idx01, idx10])
                    self.Mt01.log_and_inc("concept", "smatch_concept_sum", (self.name_idt[idx01, idx10] == 1) )
                    final_match_10[self.var2idx[1][idx10]] = self.var2idx[0][idx01]
                    quality_list10[self.var2idx[1][idx10]] = -1
                    self.Mt10.log_and_inc("concept", "bad_quality_sum", self.initial_sim[idx01, idx10])
                    self.Mt10.log_and_inc("concept", "smatch_concept_sum", (self.name_idt[idx01, idx10] == 1) )

                    key_pairs_to_delete.add((idx01, idx10))
            # the rest of them move to secondary match

            row_to_delete = []
            col_to_delete = []

            for p0, p1 in key_pairs_to_delete:

                del match_list_01[p0]
                del match_list_10[p1]

                row_to_delete.append(index_of_unmatch_01[p0])
                col_to_delete.append(index_of_unmatch_10[p1])
                unmatched_list_01.remove(p0)
                unmatched_list_10.remove(p1)

            sub_matrix = np.delete(sub_matrix, row_to_delete, axis=0)
            sub_matrix = np.delete(sub_matrix, col_to_delete, axis=1)

            if cnt > 500:
                # to select a maximum based on relation match, it also must conform with relation maximum
                raise ValueError(values+"\n\n"+str(match_list_01)+"\n\n"+str(match_list_10)+"\n\n"+self.umr0.semantic_text+"\n\n"+self.umr1.semantic_text)


        # for the rest unmatched
        for idx in match_list_01.keys():
            final_match_01[self.var2idx[0][idx]] = "NULL"+ self.var2idx[0][idx]
            quality_list01[self.var2idx[0][idx]] = -2
            self.Mt01.log_and_inc("concept", "bad_quality_sum", 0.0)
            self.Mt01.log_and_inc("concept", "smatch_concept_sum", 0.0 )
        for idx in match_list_10.keys():
            final_match_10[self.var2idx[1][idx]] = "NULL"+ self.var2idx[1][idx]
            quality_list10[self.var2idx[1][idx]] = -2
            self.Mt10.log_and_inc("concept", "bad_quality_sum", 0.0)
            self.Mt10.log_and_inc("concept", "smatch_concept_sum", 0.0)

        # assert len(final_match_01) == len(final_match_10)

        return final_match_01, final_match_10

    # @timer_decorator
    def get_new_anchor(self, final_sim):

        new_anchor = np.zeros(final_sim.shape)
        max_row = np.argmax(final_sim, axis=1)
        max_row_value = np.amax(final_sim, axis=1)
        max_col = np.argmax(final_sim, axis=0)
        max_col_value = np.amax(final_sim, axis=0)

        set01 = set()
        for i in range(new_anchor.shape[0]):
            max_pos = (final_sim[i]==max_row_value[i])
            maxima = np.where(max_pos > 0)[0]
            if np.count_nonzero(max_pos) == 1:
                set01.add((i, max_row[i]))
            else:
                tgt , wheres= Match.enhance_with_rel_label(i, maxima, 0, self.var2idx, self.umr0, self.umr1)
                if tgt:
                    set01.add((i, tgt))
        for j in range(new_anchor.shape[1]):
            max_pos = (final_sim[:, j] == max_col_value[j])
            maxima = np.where(max_pos > 0)[0]
            if np.count_nonzero(max_pos) == 1:
                if (max_col[j], j) in set01:
                    new_anchor[max_col[j], j] = 1
            else:
                tgt , wheres = Match.enhance_with_rel_label(j, maxima, 1, self.var2idx, self.umr0, self.umr1)
                if tgt:
                    if (tgt, j) in set01:
                       new_anchor[tgt, j] = 1

        return new_anchor

    def get_expedient_match(self, num, final_sim, givenv0):
        if givenv0:
            return np.argmax(final_sim, axis=1)[num]
        else:
            return np.argmax(final_sim, axis=0)[num]

    # @timer_decorator
    def iterate_anchor_and_match(
            self,
            initial_anchors,
            max_iteration = 5,
            add_asymmetric_relations = True,
            Cneighbor = 1,
            separate_1_and_2=False,
    ):
        # broadcast anchor

        # using anchors to enhance similarity information, reduce the similarity without any matched neighbors
        # anchors will broadcast, so even no anchors it would be fine

        match_list01 = {}
        quality_list01 = {}
        match_list10 = {}
        quality_list10 = {}

        anchors = initial_anchors.copy()

        c = np.where(anchors > 0)
        quality_level = 0
        for i01, i10 in zip(c[0], c[1]):
            v01 = self.var2idx[0][i01]
            v10 = self.var2idx[1][i10]
            if v01 not in match_list01.keys():
                match_list01[v01] = v10
                quality_list01[v01] = quality_level
                self.Mt01.log_and_inc("concept","good_sum", self.initial_sim[i01, i10])
                self.Mt01.log_and_inc("concept", "smatch_concept_sum", (self.name_idt[i01, i10] == 1) )
            if v10 not in match_list10.keys():
                match_list10[v10] = v01
                quality_list10[v10] = quality_level
                self.Mt10.log_and_inc("concept","good_sum", self.initial_sim[i01, i10])
                self.Mt10.log_and_inc("concept","smatch_concept_sum", (self.name_idt[i01, i10] == 1))

        for quality_level in range(1, max_iteration+1):

            old_anchors = anchors

            final_sim, broadcasted_anchor, matrices = self.incorporate_neighborhood_information(
                self.initial_sim, old_anchors, Cneighbor=Cneighbor, separate_1_and_2=separate_1_and_2,
            )

            anchors = self.get_new_anchor(final_sim)

            new_spots = anchors - old_anchors

            if np.any(new_spots < 0):                            # although a rare case, but in case old anchor vanish
                to_remove = np.where(new_spots < 0)
                for i01, i10 in zip(to_remove[0], to_remove[1]):
                    del match_list01[self.var2idx[0][i01]]
                    del match_list10[self.var2idx[1][i10]]
                    del quality_list01[self.var2idx[0][i01]]
                    del quality_list10[self.var2idx[1][i10]]

            c = np.where(anchors > 0)
            if np.any(new_spots>0):
                for i01, i10 in zip(c[0], c[1]):
                    if self.var2idx[0][i01] not in match_list01.keys():
                        match_list01[self.var2idx[0][i01]] = self.var2idx[1][i10]
                        quality_list01[self.var2idx[0][i01]] = quality_level
                        self.Mt01.log_and_inc("concept","good_sum", self.initial_sim[i01, i10])
                        self.Mt01.log_and_inc("concept", "smatch_concept_sum", (self.name_idt[i01, i10] == 1) )
                    if self.var2idx[1][i10] not in match_list10.keys():
                        match_list10[self.var2idx[1][i10]] = self.var2idx[0][i01]
                        quality_list10[self.var2idx[1][i10]] = quality_level
                        self.Mt10.log_and_inc("concept","good_sum", self.initial_sim[i01, i10])
                        self.Mt10.log_and_inc("concept", "smatch_concept_sum", (self.name_idt[i01, i10] == 1) )
            else:
                break

        unmatched_list01 = []
        unmatched_list10 = []

        if add_asymmetric_relations:
            for v in self.umr0.var2node.keys():
                if v not in match_list01.keys():
                    unmatched_list01.append(self.var2idx[0][v])

            for v in self.umr1.var2node.keys():
                if v not in match_list10.keys():
                    unmatched_list10.append(self.var2idx[1][v])

        match_list01, match_list10 = self.greedy_match_list(final_sim, unmatched_list01, unmatched_list10, match_list01, match_list10, quality_list01, quality_list10)

        return match_list01, match_list10, quality_list01, quality_list10

    # @timer_decorator
    def weighted_relational_overlap(self, sense_coefficient=0.1):

        def dictify(set2):
            rdict = defaultdict(set)
            for r in set2:
                rdict[(r[0], r[2])].add(r[1])
            return rdict

        def calculate_offsprings(edge_matrix):
            offspring_num = defaultdict(lambda: 0)
            for i in range(len(edge_matrix)):
                offspring_num[i] = ops.get_argcnt(edge_matrix, i)
            return offspring_num

        def var2name(var, vlist):
            if var in vlist.keys():
                return vlist[var].name, vlist[var].sense_id
            else:
                return var, 0

        Aout0, Aout1 = self.outmatrix

        offspring_cnt = (calculate_offsprings(Aout0) , calculate_offsprings(Aout1))

        def off_num(num, var):
            if var in self.var2idx[num].keys():
                return offspring_cnt[num][self.var2idx[num][var]]
            else:
                return 0

        def calculate_lrs(
                rel_dict_me,
                rel_dict_you,
                vlist_me,
                vlist_you,
                Mt,
                this_idx,
                match_list_me
        ):
            for (rel_pred, rel_arg), rel_label_set in rel_dict_me.items():
                # relation must be in pred-arg order
                tr_pred, tr_arg = self.translate_match(match_list_me, rel_pred), self.translate_match(match_list_me, rel_arg)
                this_sum_1 = len(rel_label_set)
                this_sum_weight_no1 = ops.rel_weight(off_num(this_idx, rel_arg), off_num(this_idx, rel_pred))   # rel_pair[1] is the argument
                if (tr_pred, tr_arg) in rel_dict_you.keys():
                    target_label_set = rel_dict_you[(tr_pred, tr_arg)]
                    tr_pred_name, tr_pred_sense = var2name(tr_pred, vlist_you)
                    tr_arg_name,  tr_arg_sense = var2name(tr_arg, vlist_you)
                    rel_pred_name, rel_pred_sense = var2name(rel_pred, vlist_me)
                    rel_arg_name, rel_arg_sense = var2name(rel_arg, vlist_me)
                    rel_overlap = ops.rel_label_set_comp(rel_label_set, target_label_set)

                    # do a test
                    rel_unlabeled_overlap = min(len(target_label_set), len(rel_label_set)) # 如果这边关系太多，肯定有不对，如果这边关系少，那肯定都对

                    rel_weighted_overlap = rel_overlap * this_sum_weight_no1
                    name_sim_pred = ops.name_sim(
                        tr_pred_name,
                        rel_pred_name,
                        (tr_pred_sense == rel_pred_sense),
                        sense_coefficient=sense_coefficient
                    )
                    name_sim_arg  = ops.name_sim(
                        tr_arg_name,
                        rel_arg_name,
                        (tr_arg_sense == rel_arg_sense),
                        sense_coefficient=sense_coefficient
                    )

                    Mt.log_and_inc("lr", "score", (name_sim_pred + name_sim_arg) / 2 * rel_overlap, this_sum_1)      # 0 是名字全对，关系全部不对的情况
                    Mt.log_and_inc("ulr", "score",(name_sim_pred + name_sim_arg) / 2 * rel_unlabeled_overlap, this_sum_1)
                    Mt.log_and_inc("wlr", "score", (name_sim_pred + name_sim_arg) / 2 * rel_weighted_overlap, this_sum_1 * this_sum_weight_no1)
                else:
                    Mt.log_and_inc("lr", "score", 0, this_sum_1)
                    Mt.log_and_inc("ulr", "score", 0, this_sum_1 )
                    Mt.log_and_inc("wlr", "score", 0, this_sum_1 * this_sum_weight_no1 )

        vlist0 = self.umr0.var2node
        vlist1 = self.umr1.var2node

        set0 = set()
        set1 = set()

        for node in vlist0.values():
            set0 = set0 | node.get_relations()
        for node in vlist1.values():
            set1 = set1 | node.get_relations()

        rel_dict0 = dictify(set0)
        rel_dict1 = dictify(set1)

        calculate_lrs(rel_dict0, rel_dict1, vlist0, vlist1, self.Mt01, 0, self.match_list01)
        calculate_lrs(rel_dict1, rel_dict0, vlist1, vlist0, self.Mt10, 0, self.match_list10)

        s0 = self.Mt01.compute("lr")
        s1 = self.Mt10.compute("lr")
        s2 = self.Mt01.compute("ulr")
        s3 = self.Mt10.compute("ulr")
        s4 = self.Mt01.compute("wlr")
        s5 = self.Mt10.compute("wlr")

        return s0, s1, ops.protected_divide(2*s0*s1, s0+s1), ops.protected_divide(2*s2*s3, s2+s3), ops.protected_divide(2*s4*s5, s4+s5)



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Main entry point for script usage"""

import argparse
import logging
import random

from ancast import params
from ancast.evaluate import evaluate

logger = logging.getLogger(__name__)


def str2bool(v):
    # https://stackoverflow.com/a/43357954
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def main():
    parser = argparse.ArgumentParser("ancast++ Argparser")

    # paths
    parser.add_argument(
        '-p', '--pred', required=True, nargs='+',
        help='(required) path to prediction file or dir '
             '(if dir, same filenames must exist in `gold`)'
    )
    parser.add_argument(
        '-pe', '--pred-ext', nargs='+', default=['txt', 'amr', 'umr', 'pred'],
        help="if `args.pred` flag is a dir, which file extensions to consider"
    )
    parser.add_argument(
        '-g', '--gold', required=True, nargs='+',
        help='(required) path to gold file or dir '
             '(if dir, same filenames must exist in `pred`)'
    )
    parser.add_argument(
        '-ge', '--gold-ext', nargs='+', default=['txt', 'amr', 'umr', 'gold'],
        help="if `args.gold` flag is a dir, which file extensions to consider"
    )
    parser.add_argument(
        '-o',  '--output',
        help='(optional) path to output analysis files'
             '(if new dir, should not contain comma in the base dirname)',
    )

    # core
    # noinspection PyTypeChecker
    parser.add_argument(
        '-s', '--scope', type=str.lower, choices=['snt', 'doc'],
        help="whether to run ancast over AMR/UMR sentence-level graphs (`snt`) "
             "or to run ancast++ over UMR document-level graphs (`doc`)"
    )
    # noinspection PyTypeChecker
    parser.add_argument(
        '-df', '--data-format', type=str.lower, choices=['amr', 'umr'],
        help='whether the input data are AMR(s) or UMR(s) '
             '(automatically set to `umr` if `--scope` is `doc`)'
    )

    # hyperparams
    parser.add_argument(
        '-c', '--cneighbor', type=int,
        default=params.CNEIGHBOR, help="coefficient for the importance of neighborhood information when broadcasting"
    )
    parser.add_argument(
        '-sc', '--sense-coefficient', type=float,
        default=params.SENSE_COEFFICIENT, help="importance of sense ID when comparing two concepts"
    )
    parser.add_argument(
        '--separate-1-and-2', type=str2bool, nargs='?', const=True,
        default=params.SEPARATE_1_AND_2, help="whether to combine one-hop and two-hop neighbors together"
    )
    parser.add_argument(
        '--allow-reify', type=str2bool, nargs='?', const=True,
        default=params.ALLOW_REIFY, help="whether to apply reification before comparing graphs"
    )
    parser.add_argument(
        '--use-alignment', type=str2bool, nargs='?', const=True,
        default=params.USE_ALIGNMENT, help="whether to use alignment information when establishing anchors"
    )
    parser.add_argument(
        '--use-smatch-top', type=str2bool, nargs='?', const=True,
        default=params.USE_SMATCH_TOP, help="whether to add (TOP :root `ROOT_NODE`) edge"
    )
    # noinspection PyTypeChecker
    ### doc-level eval specific
    parser.add_argument(
        '--weighted', type=str.lower, choices=["snts", "toks"],
        help="whether to apply weighted average for ancast++ doc-level evaluation by (1) number of sentences or (2) number of tokens"
    )

    # experiment
    parser.add_argument(
        '--seed', type=int, default=13681, help="random seed for reproducibility"
    )
    parser.add_argument(
        '--debug', action='store_true', help="debug logging mode"
    )

    args = parser.parse_args()

    # script setup
    logging.basicConfig(
        format="[%(asctime)s][%(name)s][%(levelname)s]%(message)s",
        level=logging.DEBUG if args.debug else logging.INFO
    )
    random.seed(args.seed)

    ### inner
    return evaluate(
        pred_inputs=args.pred,
        gold_inputs=args.gold,
        output_fpath=args.output,
        pred_exts=args.pred_ext,
        gold_exts=args.gold_ext,
        data_format=args.data_format,
        scope=args.scope,
        Cneighbor=args.cneighbor,
        sense_coefficient=args.sense_coefficient,
        separate_1_and_2=args.separate_1_and_2,
        allow_reify=args.allow_reify,
        use_alignment=args.use_alignment,
        use_smatch_top=args.use_smatch_top,
        weighted_average=args.weighted,
    )


if __name__ == "__main__":
    main()



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Sentence representation"""

import logging
import random
import re
from collections import defaultdict

from ancast.node import Attribute
from ancast.resource_utils import REIFY_RELS
from ancast.params import HANDLE_QUOTED_REENTRANCY
from ancast.word import Word

logger = logging.getLogger(__name__)


class Sentence:

    def __init__(
            self,
            sent,
            semantic_text,
            alignment,
            sent_num,
            format,
            # parse_tags = {"explicit"},
            allow_reify = False,
    ):
        self.text = sent
        self.invalid = False
        self.parse_tags = {"explicit"}
        self.head = None # Word()
        self.sent_num = sent_num
        self.alignment = None
        self.var2node = None # dict()
        self.umr_text = semantic_text
        self.semantic_text = semantic_text
        self.alignment = alignment # dict()
        try:
            self.head, self.var2node = self.parse(semantic_text, format)
        except Exception as e:
            print(f"Ignoring Sentence {self.sent_num} because {e} \n Content: {self.text}")
            self.invalid = True
        if allow_reify and not self.invalid:
            self.var2node = self.reify()


    def reify(self):
        """
        Reification has a semantic closeness chain:
        have-polarity-91 is the most close to the core semantic of the verb
        ``He didn't run for 5 times, so he didn't pass yesterday.''
        :cause should be added to have-polarity-91 instead of qualified;
        have-polarity-91 should also have "5 times" as lower level

        (p / pass
            :ARG0 (h / he)
            :polarity -
            :temporal (y / yesterday)
            :cause (r / run
                :ARG0 h
                :frequency 5
                :polarity -))

        actually goes to, which is

        ``That he did not run for 5 times causes his yesterday's failure to pass. ''

        (h1 / have-cause-91
            :ARG1 (ht / have-temporal-91
                :ARG1 (h2 / have-polarity-91
                    :ARG1 (p / pass
                        :ARG0 (h / he))
                    :ARG2 -)
                :ARG2 (y / yesterday))
            :ARG2 (h3 / have-polarity-91
                :ARG1 (h4 / have-frequency-91
                    :ARG1 (r / run)
                    :ARG2 5)
                :ARG2 - ))

        but it can also be:

        ``That he did not run for 5 times causes his failure, which happened yesterday.''

        (h1 / have-cause-91
            :ARG1 (h2 / have-polarity-91
                :ARG1 (p / pass
                    :ARG0 (h / he))
                :ARG2 -)
                :ARG1-of (ht / have-temporal-91
                    :ARG2 (y / yesterday)))
            :ARG2 (h3 / have-polarity-91
                :ARG1 (h4 / have-frequency-91
                    :ARG1 (r / run)
                    :ARG2 5)
                :ARG2 - ))

        But the sentence is slightly ambiguous, which can also be
        ``For 5 times, he did not run, so yesterday he failed / so he failed, which happened yesterday.''

        This is the related to semantics.

        ``He did not watch the movie for 3 hours yesterday.''
        ``He has not wached a movie for 3 years.''

        The former one, hierarchical: have-temporal-91 (yesterday) -> have-polarity-91(did not) -> have-duration-91(3 hours)
        The latter one, hierarchical: have-duration-91 (3 years) -> have-polarity-91(has not)

        Well, from James' point of view, the order of a tree matters, and the type of nodes also matter, so only events and states are concatenatable, and that all "decendant nodes" should be computed before hand.

        """

        reification_relations = REIFY_RELS
        arrived = set()
        q = [self.head]
        new_var2node = {}

        hcnt = 1

        while q:
            c = q.pop(0)
            arrived.add(c)
            new_var2node[c.var] = c
            all_relations_of_c = list(c.relations.values())
            for rs in all_relations_of_c:
                for r in rs:
                    target = r[c][0]
                    if target not in arrived:
                        if isinstance(target, Word) and (target not in q):
                            q.append(target)
                        if r.name in reification_relations.keys():
                            rconcept = "-".join(reification_relations[r.name].split("-")[:-1])
                            rsense   = int(reification_relations[r.name].split("-")[-1])
                            while ("h"+str(hcnt) in self.var2node.keys()) or ("h"+str(hcnt) in new_var2node.keys()):
                                hcnt += 1
                            new_node = Word(raw_name = rconcept, var = "h"+str(hcnt), sense_id = rsense)

                            new_node[("ARG1", {"reified", "explicit"})] = r.predicate
                            new_node[("ARG2", {"reified", "explicit"})] = r.argument

                            arrived.add(new_node)
                            r.tags.add("unreified")
                            new_var2node["h"+str(hcnt)] = new_node
                        else:
                            continue


        return new_var2node

    # @timer_decorator
    def parse(self, semantic_text, format):

        void_var = defaultdict(list)
        var2node = {}

        sense_re = re.compile(r"-[0-9]{2}")

        def parse_brackets(text, i):
            result = []

            bracket_match = False

            while i < len(text):
                if text[i] == '(':
                    cur_node, i = parse_var_content(text, i + 1)
                    result.append(cur_node)
                elif text[i] == ')':
                    i += 1
                    bracket_match = True
                    break
                else:
                    i += 1
            try:
                assert len(result) == 1, f"Multiple heads identified in semantic graph in sentence {self.sent_num}!"
            except AssertionError as error:
                print(f"Format Error: {error.args[0]}")
                raise
            try:
                assert bracket_match, f"Brackets aren't matched, please check the format in semantic graph in {self.sent_num}!"
            except AssertionError as error:
                print(f"Format Error: {error.args[0]}")
                raise

            return cur_node, i

        # @timer_decorator
        def parse_var_content(text, i):

            head = re.search(r'([a-z0-9_\-]+\s*\/\s*[^)\s]+)', text[i:]).group(1)   # 右括号或空格前所有字符
            var , txt = head.split("/")
            var = var.strip()
            txt = txt.strip()

            sense_pos_match = sense_re.search(txt)

            if sense_pos_match:
                pos = sense_pos_match.span()
                sense_id = int(txt[pos[0]+1:pos[1]])
                real_name = txt[:pos[0]]
            else:

                sense_id = 0
                real_name = txt

            i += len(head)

            this_node = Word(raw_name = real_name, var = var, sense_id = sense_id)

            try:
                assert var not in var2node, "Duplicated variable declaration, ignoring new declaration."
                var2node[var] = this_node
            except AssertionError as e:
                logger.warning(str(e))

            # early-reentrancy
            if var in void_var.keys():
                for node, rel in void_var[var]:
                    node[(rel, self.parse_tags.copy())] = this_node
                del void_var[var]

            while (i < len(text)) and (text[i] != ')'):
                if text[i] == ':':
                    voided_var = False

                    space_1 = text.find(' ', i)
                    tab_1   = text.find('\t', i)

                    end = space_1 if space_1 != -1 else tab_1
                    relation = text[i + 1:end]
                    i = end

                    while (text[i] == ' ') or (text[i] == '\t'):
                        i += 1

                    if text[i] == '(':
                        sub_node, end = parse_brackets(text, i)


                    elif text[i] == '"':

                        end = text.find('"', i+1)
                        text_part = text[i+1:end]

                        # this is an ill-formed scene where variables are quoted in some parsers

                        if HANDLE_QUOTED_REENTRANCY and (text_part in var2node.keys()):
                            sub_node = var2node[text_part]
                            print(f"Quoted reentrancy handled in sentence {self.sent_num}")
                        else:
                            sub_node = Attribute(text_part, quoted=True)
                    else:

                        firstspace = text.find(' ', i)
                        firstspace = 1e10 if firstspace == -1 else firstspace
                        firsttab = text.find('\t', i)
                        firsttab = 1e10 if firsttab == -1 else firsttab
                        firstline = text.find('\n', i)
                        firstline = 1e10 if firstline == -1 else firstline

                        next_space = min(firstline, firsttab, firstspace)
                        next_right_bracket = text.find(')', i)

                        if next_right_bracket==-1:
                            next_right_bracket = 1e10

                        end = min(next_space, next_right_bracket)

                        tt = text[i:end].strip()

                        if tt in var2node.keys():

                            # if it is a valid re-entrancy

                            sub_node = var2node[tt]

                        elif (re.fullmatch(r"s[0-9]+[a-z]+[0-9]*", tt) and (format == "umr")) or \
                             ((format == "amr") and (tt not in {"imperative", "expressive"}) and (not re.fullmatch(r"^[0-9.:+-]+$", tt))):

                            # handling of early reentrancy, where the variable is not declared yet

                            void_var[tt].append((this_node, relation))
                            voided_var = True

                        else:
                            # attributes are not quoted.
                            sub_node = Attribute(tt)

                    i = end

                    if not voided_var:
                        this_node[(relation, self.parse_tags.copy())] = sub_node

                else:
                    if re.match(r"[a-z0-9]", text[i]):
                        print(text[i-5:i+5])
                        raise RuntimeError(f"a colon is missing in {self.sent_num}.")

                    i += 1

            return this_node, i

        if len(void_var)>0:
            for v in void_var.keys():
                print(v + " is not specified!")

                # Leave the unspecified variable out for now

        return parse_brackets(semantic_text, 0)[0], var2node # [1] is "i"

    def generate_umr_text(self, random_head=False):

        if random_head:
            head = random.choice(list(self.var2node.values()))
        else:
            head = self.head

        return f"# ::id {self.sent_num}\n# ::snt {self.text}\n{ head.generate_umr(level = 0, sent_var = '', var_list = {head.var}, rel_set = set()) }\n\n"



#! /usr/bin/python3
# -*- coding: utf-8 -*-
import setuptools

setuptools.setup(
    name='ancast',
    version='0.1.0',
    description='AnCast++ metric for evaluating UMR semantic graphs.',
    # url='https://github.com/sxndqc/ancast',
    url='https://github.com/umr4nlp/ancast',
    # include_package_data=True,
    # package_data={'resources': ['*.json']},
    packages=setuptools.find_packages(exclude=['refs']),
    install_requires=[
        'numpy',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
    ],
)



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Node representation"""


class Attribute(object):

    attributes = set()

    def __init__(self, raw_name, quoted=False):

        self.raw_name = raw_name
        self.name = raw_name.lower()
        self.quoted = quoted
        Attribute.attributes.add(self.name)

    def __repr__(self):
        if self.name[0] == self.name[-1] == '"':
            return self.name[1:-1]
        else:
            return self.name

    def generate_umr(self):

        if self.quoted:
            return f'"{self.name}"'
        else:
            return self.name

# ------------ constant --------------

IMPLICIT = Attribute("implicit-argument")



#! /usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import os
from typing import Dict, List, Tuple, Union

import csv
import numpy as np

from ancast import io_utils, params
from ancast.document import DocumentMatch, SentenceMatch
from ancast.resource_utils import load_abt_vocab, load_reify_rels

logger = logging.getLogger(__name__)


def collect_fpaths_from_fdir(fpath: str, exts=None) -> List[str]:
    """depth of 1 at most,,,"""
    if os.path.isfile(fpath):
        return [fpath]

    out = []

    exts = exts or []
    exts_is_empty = len(exts) == 0
    if exts_is_empty:
        logger.warning("Received no filename extensions, collecting all files..")

    fdir = fpath
    for fname in sorted(os.listdir(fdir)):
        fpath = os.path.join(fdir, fname)
        ext = os.path.splitext(fpath)[1][1:]
        if os.path.isfile(fpath) and (ext in exts or exts_is_empty):
            out.append(fpath)
    return out

# noinspection PyShadowingBuiltins
def resolve_inputs(
        inputs,
        exts, delimiter="\n\n",
        return_fnames=False
) -> Union[List[str], Tuple[List[str], List[str]]]:
    if isinstance(inputs, str):
        inputs = [inputs]

    # load strings from fpaths
    fnames = []  # without extension
    inputs_resolved = []
    for input in inputs:
        if os.path.exists(input):
            # is a fpath, should be loaded
            pred_input_fpaths_list = collect_fpaths_from_fdir(input, exts=exts)

            for pred_input_fpath in pred_input_fpaths_list:
                inputs_resolved.append(
                    io_utils.load_txt(pred_input_fpath, delimiter=delimiter)
                )
                fnames.append(os.path.splitext(os.path.basename(pred_input_fpath))[0])
        else:
            # otherwise, should be a string input ready for eval
            inputs_resolved.append([input])
            fnames.append(None)

    if return_fnames:
        return inputs_resolved, fnames
    return inputs_resolved


# noinspection PyPep8Naming
def evaluate_snt(
        pred_inputs: Union[str, List[str]],
        gold_inputs: Union[str, List[str]],
        data_format: str,
        output_fpath=None,
        pred_exts=None,
        gold_exts=None,
        Cneighbor: int = params.CNEIGHBOR,
        sense_coefficient: float = params.SENSE_COEFFICIENT,
        separate_1_and_2: bool = params.SEPARATE_1_AND_2,
        allow_reify: bool = params.ALLOW_REIFY,
        use_alignment: bool = params.USE_ALIGNMENT,
        use_smatch_top: bool = params.USE_SMATCH_TOP,
        **unused
) -> List[float]:
    # sentinel
    data_format = data_format or "umr"
    assert data_format in ['amr', 'umr'], \
        f"expected `amr` or `umr` data format, but received `{data_format}`"
    delimiter = "\n\n" if data_format == "amr" else "\n\n\n"

    # maybe load external resources
    if allow_reify:
        logger.debug("Loading external resources..")
        load_abt_vocab(global_cache=True)
        load_reify_rels(global_cache=True)

    pred_inputs_resolved, pred_fnames = resolve_inputs(
        pred_inputs, exts=pred_exts, delimiter=delimiter, return_fnames=True
    )
    gold_inputs_resolved = resolve_inputs(
        gold_inputs, exts=gold_exts, delimiter=delimiter
    )
    num_pred_inputs = len(pred_inputs_resolved)
    num_gold_inputs = len(gold_inputs_resolved)

    # there is no guarantee that the number of pred and gold inputs match perfectly
    if num_pred_inputs != num_gold_inputs:
        logger.warning(
            "Number of pred and gold inputs do not match (%d vs %d)",
            num_pred_inputs, num_gold_inputs
        )

    output_flag = output_fpath is not None

    # just iterate through pred inputs
    fscores = []
    for i, (pred_input_blocks, pred_fname) in enumerate(
            zip(pred_inputs_resolved, pred_fnames)):
        logger.debug("Current Pred File: `%s`", pred_fname)

        if i >= num_gold_inputs:
            logger.warning(
                "Found no matching gold input for `%s`, skipping..",
                pred_fname
            )
            continue

        csv_f = csv_writer = None
        if output_flag:
            write_title = True

            output_is_dir = os.path.isdir(output_fpath)
            if output_is_dir:
                pred_output = os.path.join(output_fpath, f'{pred_fname}.csv')
                csv_f = open(pred_output, 'w', newline='')
            else:
                pred_output = output_fpath
                if io_utils.csv_is_empty(pred_output):
                    write_title = False
                csv_f = open(pred_output, 'a', newline='')

            csv_writer = csv.writer(csv_f)
            if write_title:
                csv_writer.writerow(
                    [
                        "Sentence", data_format + "_test", data_format + "_gold",
                        "Match T2G", "Match G2T",
                        "Concept Match Precision", "Concept Match Recall", "Concept Match F Score",
                        "Labeled Relational Match Precision", "Labeled Relational Match Recall",
                        "LRM F-Score", "ULRM F-Score", "WLRM F-score", "Smatch Score"
                    ]
                )

        match = SentenceMatch(
            format=data_format,
            Cneighbor=Cneighbor,
            sense_coefficient=sense_coefficient,
            allow_reify=allow_reify,
            separate_1_and_2=separate_1_and_2,
            use_alignment=use_alignment,
            use_smatch_top=use_smatch_top,
        )

        # compute fscores while accumulating messages for analysis
        fscore = match.compute_scores(
            pred_inputs=pred_input_blocks,
            gold_inputs=gold_inputs_resolved[i]
        )
        fscores.append(fscore)

        if output_flag:
            for msg in match.msgs:
                csv_writer.writerow(msg)
            csv_f.close()

    return fscores

def evaluate_doc(
        pred_inputs: Union[str, List[str]],
        gold_inputs: Union[str, List[str]],
        output_fpath=None,
        pred_exts=None,
        gold_exts=None,
        Cneighbor: int = params.CNEIGHBOR,
        sense_coefficient: float = params.SENSE_COEFFICIENT,
        separate_1_and_2: bool = params.SEPARATE_1_AND_2,
        allow_reify: bool = params.ALLOW_REIFY,
        use_alignment: bool = params.USE_ALIGNMENT,
        use_smatch_top: bool = params.USE_SMATCH_TOP,
        weighted=None,
        **unused
) -> Dict[str, Union[float, list[float]]]:
    # maybe load external resources
    if allow_reify:
        logger.debug("Loading external resources..")
        load_abt_vocab(global_cache=True)
        load_reify_rels(global_cache=True)

    pred_inputs_resolved, pred_fnames = resolve_inputs(
        pred_inputs, exts=pred_exts, delimiter="\n\n\n", return_fnames=True
    )
    gold_inputs_resolved = resolve_inputs(
        gold_inputs, exts=gold_exts, delimiter="\n\n\n"
    )
    num_pred_inputs = len(pred_inputs_resolved)
    num_gold_inputs = len(gold_inputs_resolved)

    # there is no guarantee that the number of pred and gold inputs match perfectly
    if num_pred_inputs != num_gold_inputs:
        logger.warning(
            "Number of pred and gold inputs do not match (%d vs %d)",
            num_pred_inputs, num_gold_inputs
        )

    output_flag = output_fpath is not None

    aggregate_flag = num_pred_inputs > 1
    aggregate = {
        'sent': [],
        'modal': [],
        'temporal': [],
        'coref': [],
        'comp': []
    }

    num_snts_by_docs, num_toks_by_docs = [], []
    for i, (pred_input_blocks, pred_fname) in enumerate(
            zip(pred_inputs_resolved, pred_fnames)):
        logger.debug("Current Pred File: `%s`", pred_fname)

        if i >= num_gold_inputs:
            logger.warning(
                "Found no matching gold input for `%s`, skipping..",
                pred_fname
            )
            continue

        csv_f = csv_writer = None
        if output_flag:
            write_title = True

            output_is_dir = os.path.isdir(output_fpath)
            if output_is_dir:
                pred_output = os.path.join(output_fpath, f'{pred_fname}.csv')
                csv_f = open(pred_output, 'w', newline='')
            else:
                pred_output = output_fpath
                if io_utils.csv_is_empty(pred_output):
                    write_title = False
                csv_f = open(pred_output, 'a', newline='')

            csv_writer = csv.writer(csv_f)
            if write_title:
                csv_writer.writerow(
                    [
                        "Sentence", "umr_test", "umr_gold",
                        "Match T2G", "Match G2T",
                        "Concept Match Precision", "Concept Match Recall", "Concept Match F Score",
                        "Labeled Relational Match Precision", "Labeled Relational Match Recall",
                        "LRM F-Score", "ULRM F-Score", "WLRM F-score", "Smatch Score"
                    ]
                )

        match = DocumentMatch(
            Cneighbor=Cneighbor,
            sense_coefficient=sense_coefficient,
            allow_reify=allow_reify,
            separate_1_and_2=separate_1_and_2,
            use_alignment=use_alignment,
            use_smatch_top=use_smatch_top,
        )

        # compute fscores while accumulating messages for analysis
        match.compute_scores(
            pred_inputs=pred_input_blocks,
            gold_inputs=gold_inputs_resolved[i]
        )
        for k,v in aggregate.items():
            v.append(getattr(match, f'{k}_fscore'))

        num_snts_by_docs.append(match.doc_num_snts)
        num_toks_by_docs.append(match.doc_num_toks)

        if output_flag:
            for msg in match.msgs:
                csv_writer.writerow(msg)
            csv_f.close()

    if aggregate_flag:
        logger.info("=== Aggregate Statistics ===")

        apply_weighted_average = weighted is not None
        for k,v in aggregate.items():
            weights = None
            if apply_weighted_average:
                weights = num_snts_by_docs if weighted == 'snts' else num_toks_by_docs

            aggregate[k] = cur_aggr = np.average(v, weights=weights)
            logger.info(f"{'Weighted ' if weighted else ''}Aggregate {k.capitalize()} Score:\t{cur_aggr:.2%}")

    else:
        for k, v in aggregate.items():
            aggregate[k] = v[0]

    return aggregate

# @timer_decorator
def evaluate(*args, scope: str = "doc", **kwargs):
    # sentinel
    scope = scope.lower()
    assert scope in ['snt', 'doc'], \
        f'either `snt` or `doc` scope expected but received `{scope}`'

    eval_fn = evaluate_doc if scope == 'doc' else evaluate_snt
    return eval_fn(*args, **kwargs)



#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""misc. utils"""

import codecs
import json
import logging
import os

logger = logging.getLogger(__name__)


### I/O
def get_fpath(fpath_or_dir, fname=None):
  fpath = fpath_or_dir
  if fname is not None:
    if not os.path.isdir(fpath_or_dir):
      fpath_or_dir = os.path.dirname(fpath_or_dir)
    fpath = os.path.join(fpath_or_dir, fname)
  return fpath

# txt
def load_txt(fpath_or_dir, fname=None, encoding=None, delimiter=None):
  fpath = get_fpath(fpath_or_dir, fname)
  with codecs.open(fpath, encoding=encoding, errors='ignore') as f:
    data = f.read()
  if delimiter is not None:  # drop empty strings
    data = list(filter(None, data.split(delimiter)))
  return data

def save_txt(obj, fpath_or_dir, fname=None, delimiter=None):
  fpath = get_fpath(fpath_or_dir, fname)
  with open(fpath, 'w') as f:
    if delimiter is not None:
      obj = delimiter.join(obj)
    f.write(obj)
  return fpath

# json
def load_json(fpath_or_dir, fname=None):
  fpath = get_fpath(fpath_or_dir, fname)
  with open(fpath) as f:
    obj = json.loads(f.read())
  return obj

def save_json(obj, fpath_or_dir, fname=None, indent=4):
  fpath = get_fpath(fpath_or_dir, fname)
  with open(fpath, 'w') as f:
    json.dump(obj, f, indent=indent)
  return fpath

# csv
def csv_is_empty(fpath):
  if os.path.exists(fpath) and os.path.getsize(fpath) > 0:
    with open(fpath, 'r', encoding='utf-8') as f:
      first_line = f.readline()
      return first_line.strip() == ''
  return True

